@@ -25,6 +25,10 @@
 #endif
 
 #include <libavcodec/avcodec.h>
+#include <libavutil/imgutils.h>
+#include <libavfilter/avfilter.h>
+#include <libavfilter/buffersrc.h>
+#include <libavfilter/buffersink.h>
 
 #include <gst/gst.h>
 #include <gst/video/video.h>
@@ -91,8 +95,16 @@ typedef struct _GstFFMpegDeinterlace
   gboolean reconfigure;
   GstFFMpegDeinterlaceMode new_mode;
 
-  enum PixelFormat pixfmt;
-  AVPicture from_frame, to_frame;
+  enum AVPixelFormat pixfmt;
+  AVFrame from_frame, to_frame;
+
+  AVFilterContext *buffersink_ctx;
+  AVFilterContext *buffersrc_ctx;
+  AVFilterGraph *filter_graph;
+  AVFrame *filter_frame;
+  int last_width, last_height;
+  enum AVPixelFormat last_pixfmt;
+
 } GstFFMpegDeinterlace;
 
 typedef struct _GstFFMpegDeinterlaceClass
@@ -135,6 +147,8 @@ G_DEFINE_TYPE (GstFFMpegDeinterlace, gst_ffmpegdeinterlace, GST_TYPE_ELEMENT);
 static GstFlowReturn gst_ffmpegdeinterlace_chain (GstPad * pad,
     GstObject * parent, GstBuffer * inbuf);
 
+static void gst_ffmpegdeinterlace_dispose (GObject * obj);
+
 static void
 gst_ffmpegdeinterlace_class_init (GstFFMpegDeinterlaceClass * klass)
 {
@@ -159,14 +173,16 @@ gst_ffmpegdeinterlace_class_init (GstFFMpegDeinterlaceClass * klass)
           DEFAULT_MODE, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS)
       );
 
-  gst_element_class_add_pad_template (element_class,
-      gst_static_pad_template_get (&src_factory));
-  gst_element_class_add_pad_template (element_class,
-      gst_static_pad_template_get (&sink_factory));
+  gst_element_class_add_static_pad_template (element_class, &src_factory);
+  gst_element_class_add_static_pad_template (element_class, &sink_factory);
 
   gst_element_class_set_static_metadata (element_class,
       "libav Deinterlace element", "Filter/Effect/Video/Deinterlace",
       "Deinterlace video", "Luca Ognibene <luogni@tin.it>");
+
+  gobject_class->dispose = gst_ffmpegdeinterlace_dispose;
+
+  gst_type_mark_as_plugin_api (GST_TYPE_FFMPEGDEINTERLACE_MODES, 0);
 }
 
 static void
@@ -180,10 +196,10 @@ gst_ffmpegdeinterlace_update_passthrough (GstFFMpegDeinterlace * deinterlace)
 }
 
 static gboolean
-gst_ffmpegdeinterlace_sink_setcaps (GstPad * pad, GstCaps * caps)
+gst_ffmpegdeinterlace_sink_setcaps (GstPad * pad, GstObject * parent,
+    GstCaps * caps)
 {
-  GstFFMpegDeinterlace *deinterlace =
-      GST_FFMPEGDEINTERLACE (gst_pad_get_parent (pad));
+  GstFFMpegDeinterlace *deinterlace = GST_FFMPEGDEINTERLACE (parent);
   GstStructure *structure = gst_caps_get_structure (caps, 0);
   const gchar *imode;
   AVCodecContext *ctx;
@@ -209,18 +225,17 @@ gst_ffmpegdeinterlace_sink_setcaps (GstPad * pad, GstCaps * caps)
   ctx->pix_fmt = AV_PIX_FMT_NB;
   gst_ffmpeg_caps_with_codectype (AVMEDIA_TYPE_VIDEO, caps, ctx);
   if (ctx->pix_fmt == AV_PIX_FMT_NB) {
-    gst_ffmpeg_avcodec_close (ctx);
-    av_free (ctx);
+    avcodec_free_context (&ctx);
     return FALSE;
   }
 
   deinterlace->pixfmt = ctx->pix_fmt;
 
-  av_free (ctx);
+  avcodec_free_context (&ctx);
 
   deinterlace->to_size =
-      avpicture_get_size (deinterlace->pixfmt, deinterlace->width,
-      deinterlace->height);
+      av_image_get_buffer_size (deinterlace->pixfmt, deinterlace->width,
+      deinterlace->height, 1);
 
   src_caps = gst_caps_copy (caps);
   gst_caps_set_simple (src_caps, "interlace-mode", G_TYPE_STRING,
@@ -244,7 +259,7 @@ gst_ffmpegdeinterlace_sink_event (GstPad * pad, GstObject * parent,
       GstCaps *caps;
 
       gst_event_parse_caps (event, &caps);
-      ret = gst_ffmpegdeinterlace_sink_setcaps (pad, caps);
+      ret = gst_ffmpegdeinterlace_sink_setcaps (pad, parent, caps);
       gst_event_unref (event);
       break;
     }
@@ -277,6 +292,105 @@ gst_ffmpegdeinterlace_init (GstFFMpegDeinterlace * deinterlace)
   deinterlace->reconfigure = FALSE;
   deinterlace->mode = DEFAULT_MODE;
   deinterlace->new_mode = -1;
+  deinterlace->last_width = -1;
+  deinterlace->last_height = -1;
+  deinterlace->last_pixfmt = AV_PIX_FMT_NONE;
+}
+
+static void
+delete_filter_graph (GstFFMpegDeinterlace * deinterlace)
+{
+  if (deinterlace->filter_graph) {
+    av_frame_free (&deinterlace->filter_frame);
+    avfilter_graph_free (&deinterlace->filter_graph);
+  }
+}
+
+static void
+gst_ffmpegdeinterlace_dispose (GObject * obj)
+{
+  GstFFMpegDeinterlace *deinterlace = GST_FFMPEGDEINTERLACE (obj);
+
+  delete_filter_graph (deinterlace);
+
+  G_OBJECT_CLASS (gst_ffmpegdeinterlace_parent_class)->dispose (obj);
+}
+
+static int
+init_filter_graph (GstFFMpegDeinterlace * deinterlace,
+    enum AVPixelFormat pixfmt, int width, int height)
+{
+  AVFilterInOut *inputs = NULL, *outputs = NULL;
+  char args[512];
+  int res;
+
+  delete_filter_graph (deinterlace);
+  deinterlace->filter_graph = avfilter_graph_alloc ();
+  snprintf (args, sizeof (args),
+      "buffer=video_size=%dx%d:pix_fmt=%d:time_base=1/1:pixel_aspect=0/1[in];"
+      "[in]yadif[out];" "[out]buffersink", width, height, pixfmt);
+  res =
+      avfilter_graph_parse2 (deinterlace->filter_graph, args, &inputs,
+      &outputs);
+  if (res < 0)
+    return res;
+  if (inputs || outputs)
+    return -1;
+  res = avfilter_graph_config (deinterlace->filter_graph, NULL);
+  if (res < 0)
+    return res;
+
+  deinterlace->buffersrc_ctx =
+      avfilter_graph_get_filter (deinterlace->filter_graph, "Parsed_buffer_0");
+  deinterlace->buffersink_ctx =
+      avfilter_graph_get_filter (deinterlace->filter_graph,
+      "Parsed_buffersink_2");
+  if (!deinterlace->buffersrc_ctx || !deinterlace->buffersink_ctx)
+    return -1;
+  deinterlace->filter_frame = av_frame_alloc ();
+  deinterlace->last_width = width;
+  deinterlace->last_height = height;
+  deinterlace->last_pixfmt = pixfmt;
+
+  return 0;
+}
+
+static int
+process_filter_graph (GstFFMpegDeinterlace * deinterlace, AVFrame * dst,
+    const AVFrame * src, enum AVPixelFormat pixfmt, int width, int height)
+{
+  int res;
+
+  if (!deinterlace->filter_graph || width != deinterlace->last_width ||
+      height != deinterlace->last_height
+      || pixfmt != deinterlace->last_pixfmt) {
+    res = init_filter_graph (deinterlace, pixfmt, width, height);
+    if (res < 0)
+      return res;
+  }
+
+  memcpy (deinterlace->filter_frame->data, src->data, sizeof (src->data));
+  memcpy (deinterlace->filter_frame->linesize, src->linesize,
+      sizeof (src->linesize));
+  deinterlace->filter_frame->width = width;
+  deinterlace->filter_frame->height = height;
+  deinterlace->filter_frame->format = pixfmt;
+  res =
+      av_buffersrc_add_frame (deinterlace->buffersrc_ctx,
+      deinterlace->filter_frame);
+  if (res < 0)
+    return res;
+  res =
+      av_buffersink_get_frame (deinterlace->buffersink_ctx,
+      deinterlace->filter_frame);
+  if (res < 0)
+    return res;
+  av_image_copy (dst->data, dst->linesize,
+      (const uint8_t **) deinterlace->filter_frame->data,
+      deinterlace->filter_frame->linesize, pixfmt, width, height);
+  av_frame_unref (deinterlace->filter_frame);
+
+  return 0;
 }
 
 static GstFlowReturn
@@ -290,17 +404,16 @@ gst_ffmpegdeinterlace_chain (GstPad * pad, GstObject * parent,
 
   GST_OBJECT_LOCK (deinterlace);
   if (deinterlace->reconfigure) {
+    GstCaps *caps;
+
     if ((gint) deinterlace->new_mode != -1)
       deinterlace->mode = deinterlace->new_mode;
     deinterlace->new_mode = -1;
 
     deinterlace->reconfigure = FALSE;
     GST_OBJECT_UNLOCK (deinterlace);
-    if (gst_pad_has_current_caps (deinterlace->srcpad)) {
-      GstCaps *caps;
-
-      caps = gst_pad_get_current_caps (deinterlace->sinkpad);
-      gst_ffmpegdeinterlace_sink_setcaps (deinterlace->sinkpad, caps);
+    if ((caps = gst_pad_get_current_caps (deinterlace->srcpad))) {
+      gst_ffmpegdeinterlace_sink_setcaps (deinterlace->sinkpad, parent, caps);
       gst_caps_unref (caps);
     }
   } else {
@@ -320,8 +433,9 @@ gst_ffmpegdeinterlace_chain (GstPad * pad, GstObject * parent,
   gst_ffmpeg_avpicture_fill (&deinterlace->to_frame, to_map.data,
       deinterlace->pixfmt, deinterlace->width, deinterlace->height);
 
-  avpicture_deinterlace (&deinterlace->to_frame, &deinterlace->from_frame,
-      deinterlace->pixfmt, deinterlace->width, deinterlace->height);
+  process_filter_graph (deinterlace, &deinterlace->to_frame,
+      &deinterlace->from_frame, deinterlace->pixfmt, deinterlace->width,
+      deinterlace->height);
   gst_buffer_unmap (outbuf, &to_map);
   gst_buffer_unmap (inbuf, &from_map);
 
